/* This is a module which is used for setting the TOS field of a packet. */

/* (C) 1999-2001 Paul `Rusty' Russell
 * (C) 2002-2004 Netfilter Core Team <coreteam@xxxxxxxxxxxxx>
 * © 2007 CC Computer Consultants GmbH <jengelh@xxxxxxxxxxxxxxx>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <net/checksum.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_ipv4/ipt_TOS.h>

static unsigned int
xt_TOS_target4(const struct sk_buff **pskb,
            const struct net_device *in,
            const struct net_device *out,
            unsigned int hooknum,
            const struct xt_target *target,
            const void *targinfo)
{
	const struct ipt_tos_target_info *info = targinfo;
	u_int8_t old_tos;

	if (((*pskb)->nh.iph->tos & IPTOS_TOS_MASK) == info->tos)
		return XT_CONTINUE;

    u_int16_t diffs[2];

	if (!skb_make_writable(pskb, sizeof(struct iphdr)))
		return NF_DROP;

    diffs[0] = htons((*pskb)->nh.iph->tos) ^ 0xFFFF;
	(*pskb)->nh.iph->tos = ((*pskb)->nh.iph->tos & IPTOS_PREC_MASK)
	           | info->tos;
    diffs[1] = htons((*pskb)->nh.iph->tos);
	(*pskb)->nh.iph->check
			= csum_fold(csum_partial((char *)diffs,
						 sizeof(diffs),
						 (*pskb)->nh.iph->check
						 ^0xFFFF));
	return XT_CONTINUE;
}

static unsigned int
xt_TOS_target6(const struct sk_buff **pskb,
            const struct net_device *in,
            const struct net_device *out,
            unsigned int hooknum,
            const struct xt_target *target,
            const void *targinfo)
{
	const struct ipt_tos_target_info *info = targinfo;

	if ((*pskb)->nh.ipv6h->priority == info->tos)
		return XT_CONTINUE;

	if (!skb_make_writable(pskb, sizeof(struct ipv6hdr)))
		return NF_DROP;

	(*pskb)->nh.ipv6h->priority ^= ((info->tos) >> 4);
	(*pskb)->nh.ipv6h->flow_lbl[0] ^= ((info->tos) << 4);
	return XT_CONTINUE;
}

static int
xt_TOS_check4(const char *tablename,
		      const void *entry,
		      const struct xt_target *target,
		      void *targinfo,
		      unsigned int hook_mask)
{
	const struct ipt_tos_target_info *info = targinfo;

	if ((info->tos & ~IPTOS_TOS_MASK) != 0) {
		printk(KERN_WARNING KBUILD_MODNAME
			   ": Bad TOS value %#x\n", info->tos);
		return false;
	}

	return true;
}

static int
xt_TOS_check6(const char *tablename,
		      const void *entry,
		      const struct xt_target *target,
		      void *targinfo,
		      unsigned int hook_mask)
{
	const struct ipt_tos_target_info *info = targinfo;

	if (info->tos > 0x10) {
		printk(KERN_WARNING KBUILD_MODNAME
		       ": Traffic Class field may only take values 0-16\n");
		return false;
	}

	return true;
}

static struct xt_target xt_TOS_reg[] = {
	{
		.name		= "TOS",
		.family		= AF_INET,
		.target		= xt_TOS_target4,
		.targetsize	= sizeof(struct ipt_tos_target_info),
		.table		= "mangle",
		.checkentry	= xt_TOS_check4,
		.me		= THIS_MODULE,
	},
	{
		.name		= "TOS",
		.family		= AF_INET6,
		.target		= xt_TOS_target6,
		.targetsize	= sizeof(struct ipt_tos_target_info),
		.table		= "mangle",
		.checkentry	= xt_TOS_check6,
		.me		= THIS_MODULE,
	},
};

static int __init xt_TOS_init(void)
{
	return xt_register_targets(xt_TOS_reg, ARRAY_SIZE(xt_TOS_reg));
}

static void __exit xt_TOS_exit(void)
{
	xt_unregister_targets(xt_TOS_reg, ARRAY_SIZE(xt_TOS_reg));
}

module_init(xt_TOS_init);
module_exit(xt_TOS_exit);
MODULE_AUTHOR("Netfilter Core Team <coreteam@xxxxxxxxxxxxx>");
MODULE_DESCRIPTION("netfilter \"TOS\" target module");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_TOS");
MODULE_ALIAS("ip6t_TOS");




